from typing import Dict

from pydantic import BaseModel

from watchmen_model.common import TenantId
from watchmen_model.system import tenant


class LineageGraphs(BaseModel):
	directed:bool = None
	multigraph:bool = None
	graph:Dict = {}
	tenantId : TenantId = None


class LineageNodes(BaseModel):
	id:str = None
	name:str = None
	pass


class LineageLinks(BaseModel):



	pass


from typing import Optional, Union

from pydantic import BaseModel

from watchmen_model.common import UserBasedTuple, Auditable, LastVisit
from watchmen_model.common.tuple_ids import ObjectiveReportId, DerivedObjectiveReportId
from watchmen_model.indicator.objective_report import ObjectiveReport


def construct_objective_report(definition: Optional[Union[dict, ObjectiveReport]]) -> Optional[ObjectiveReport]:
	if definition is None:
		return None
	elif isinstance(definition, ObjectiveReport):
		return definition
	else:
		# noinspection PyArgumentList
		return ObjectiveReport(**definition)

class DerivedObjectiveReport(UserBasedTuple, Auditable, LastVisit, BaseModel):
	derivedObjectiveReportId: DerivedObjectiveReportId = None
	name: str = None
	description: str = None
	objectiveReportId: ObjectiveReportId
	definition: ObjectiveReport

	def __setattr__(self, name, value):
		if name == 'definition':
			super().__setattr__(name, construct_objective_report(value))
		else:
			super().__setattr__(name, value)


from typing import Dict, List, Optional, Union

from _decimal import Decimal
from pydantic import BaseModel

from watchmen_model.common import Auditable, LastVisit, UserBasedTuple
from watchmen_model.common.tuple_ids import ObjectiveReportId
from watchmen_model.indicator import ObjectiveTarget, ObjectiveTimeFrame, ObjectiveVariable, ObjectiveVariableKind, \
	ObjectiveVariableOnBucket, ObjectiveVariableOnRange, ObjectiveVariableOnValue
from watchmen_utilities import ArrayHelper


class Variable(BaseModel):
	name: str = None
	value: str = None


class CellValue(BaseModel):
	currentValue: Decimal = None
	previousValue: Decimal = None
	chainValue: Decimal = None
	failed: bool = False


class CellTarget(BaseModel):
	name: str = None
	row: str = None
	cell: str = None
	objectiveId: str = None
	value: CellValue = CellValue()
	targetId: str = None
	parameters: Dict = None


def construct_cells(cells: Optional[list] = None) -> Optional[List[CellTarget]]:
	if cells is None:
		return None
	else:
		return ArrayHelper(cells).map(lambda x: construct_cell(x)).to_list()


def construct_target(target: Optional[Union[dict, ObjectiveTarget]]) -> Optional[ObjectiveTarget]:
	if target is None:
		return None
	elif isinstance(target, ObjectiveTarget):
		return target
	else:
		# noinspection PyArgumentList
		return ObjectiveTarget(**target)


def construct_cell(cell: Optional[dict] = None) -> Optional[CellTarget]:
	if cell is None:
		return None
	else:
		return CellTarget(**cell)


def construct_targets(targets: Optional[list] = None) -> Optional[List[ObjectiveTarget]]:
	if targets is None:
		return None
	else:
		return ArrayHelper(targets).map(lambda x: construct_target(x)).to_list()


def construct_variable(variable: Optional[Union[dict, ObjectiveVariable]]) -> Optional[ObjectiveVariable]:
	if variable is None:
		return None
	elif isinstance(variable, ObjectiveVariable):
		return variable
	else:
		kind = variable.get('kind')
		if kind == ObjectiveVariableKind.SINGLE_VALUE:
			return ObjectiveVariableOnValue(**variable)
		elif kind == ObjectiveVariableKind.RANGE:
			return ObjectiveVariableOnRange(**variable)
		elif kind == ObjectiveVariableKind.BUCKET:
			return ObjectiveVariableOnBucket(**variable)
		else:
			raise Exception(f'Objective variable kind[{kind}] cannot be recognized.')


def construct_variables(variables: Optional[list] = None) -> Optional[List[ObjectiveVariable]]:
	if variables is None:
		return None
	else:
		return ArrayHelper(variables).map(lambda x: construct_variable(x)).to_list()


class ObjectiveReport(UserBasedTuple, Auditable, LastVisit, BaseModel):
	objectiveReportId: ObjectiveReportId = None
	name: str = None
	variables: List[ObjectiveVariable] = []
	timeFrame: ObjectiveTimeFrame = []
	cells: List[CellTarget] = []

	def __setattr__(self, name, value):
		if name == 'cells':
			super().__setattr__(name, construct_cells(value))
		elif name == 'targets':
			super().__setattr__(name, construct_targets(value))
		elif name == 'variables':
			super().__setattr__(name, construct_variables(value))
		else:
			super().__setattr__(name, value)



from typing import List

from watchmen_model.admin import Topic, TopicKind, TopicType, Factor, FactorType, FactorIndexGroup, Pipeline


def ask_query_performance_topics():
	return [
		Topic(
			name='query_performance_log',
			kind=TopicKind.SYSTEM,
			type=TopicType.RAW,
			factors=[
				Factor(factorId='rpml-f-1', name='uid', type=FactorType.TEXT),
				Factor(
					factorId='rpml-f-2', name='topic_dimensions', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_1, precision='200'),
				Factor(
					factorId='rpml-f-3', name='column_dimensions', type=FactorType.TEXT,
					flatten=True, indexGroup=FactorIndexGroup.INDEX_2, precision='200'),
				Factor(
					factorId='rpml-f-4', name='execution_time', type=FactorType.NUMBER, precision='50'),
				Factor(
					factorId='rpml-f-5', name='data_volume', type=FactorType.NUMBER, precision='50'),
				Factor(
					factorId='rpml-f-6', name='join_dimensions', type=FactorType.TEXT,
					flatten=True, precision='200'),
				Factor(
					factorId='rpml-f-7', name='where_dimensions', type=FactorType.TEXT,
					flatten=True, precision='200'),
				Factor(
					factorId='rpml-f-8', name='group_by_dimensions', type=FactorType.TEXT,
					flatten=True, precision='200'),

			],
			description='query performance log raw topic ',
		)
	]


def ask_query_performance_pipelines(topics: List[Topic]) -> List[Pipeline]:
	# TODO define all pipeline monitor pipelines
	return []



from enum import Enum
from typing import Dict

from watchmen_model.common import TenantBasedTuple


class Operation(TenantBasedTuple):
	recordId: str
	operationType: str
	tupleKey: str
	tupleType: str
	tupleId: str
	content: Dict
	versionNum: str


class OperationType(str, Enum):
	CREATE = "create"
	UPDATE = "update"
	DELETE = "delete"



from typing import TypeVar

TenantId = TypeVar('TenantId', bound=str)

DataSourceId = TypeVar('DataSourceId', bound=str)
ExternalWriterId = TypeVar('ExternalWriterId', bound=str)
PluginId = TypeVar('PluginId', bound=str)

UserId = TypeVar('UserId', bound=str)
UserGroupId = TypeVar('UserGroupId', bound=str)

EnumId = TypeVar('EnumId', bound=str)
EnumItemId = TypeVar('EnumItemId', bound=str)
FactorId = TypeVar('FactorId', bound=str)
TopicId = TypeVar('TopicId', bound=str)
PipelineActionId = TypeVar('PipelineActionId', bound=str)
PipelineUnitId = TypeVar('PipelineUnitId', bound=str)
PipelineStageId = TypeVar('PipelineStageId', bound=str)
PipelineId = TypeVar('PipelineId', bound=str)
PipelineGraphicId = TypeVar('PipelineGraphicId', bound=str)

ReportFunnelId = TypeVar('ReportFunnelId', bound=str)
ReportId = TypeVar('ReportId', bound=str)
SubjectDatasetColumnId = TypeVar('SubjectDatasetColumnId', bound=str)
SubjectId = TypeVar('SubjectId', bound=str)
SpaceId = TypeVar('SpaceId', bound=str)
ConnectedSpaceId = TypeVar('ConnectedSpaceId', bound=str)
DashboardId = TypeVar('DashboardId', bound=str)

BucketId = TypeVar('BucketId', bound=str)
IndicatorId = TypeVar('IndicatorId', bound=str)
ObjectiveId = TypeVar('ObjectiveId', bound=str)
ObjectiveReportId = TypeVar('ObjectiveReportId', bound=str)
DerivedObjectiveReportId = TypeVar('DerivedObjectiveReportId', bound=str)
ObjectiveFactorId = TypeVar('ObjectiveFactorId', bound=str)
ObjectiveTargetId = TypeVar('ObjectiveTargetId', bound=str)
DerivedObjectiveId = TypeVar('DerivedObjectiveId', bound=str)
BreakdownTargetId = TypeVar('BreakdownTargetId', bound=str)
AchievementPluginTaskId = TypeVar('AchievementPluginTaskId', bound=str)
ConvergenceId = TypeVar('ConvergenceId', bound=str)
ConvergenceVariableId = TypeVar('ConvergenceVariableId', bound=str)
ConvergenceTargetVariableMappingId = TypeVar('ConvergenceTargetVariableMappingId', bound=str)
ConvergenceTargetId = TypeVar('ConvergenceTargetId', bound=str)

PatId = TypeVar('PatId', bound=str)

EventDefinitionId = TypeVar('EventDefinitionId', bound=str)
NotificationDefinitionId = TypeVar('NotificationDefinitionId', bound=str)
SubscriptionEventId = TypeVar('SubscriptionEventId', bound=str)
SubscriptionEventLockId = TypeVar('SubscriptionEventLockId', bound=str)

CompetitiveLockId = TypeVar('CompetitiveLockId', bound=int)
ScheduledTaskId = TypeVar('ScheduledTaskId', bound=int)
ChangeRecordId = TypeVar('ChangeRecordId', bound=int)
ChangeJsonId = TypeVar('ChangeJsonId', bound=int)
CollectorModelConfigId = TypeVar('CollectorModelConfigId', bound=str)
CollectorTableConfigId = TypeVar('CollectorTableConfigId', bound=str)
EventTriggerId = TypeVar('EventTriggerId', bound=int)
ModelTriggerId = TypeVar('ModelTriggerId', bound=int)
TableTriggerId = TypeVar('TableTriggerId', bound=int)



from typing import Dict, Any, List

from pydantic import BaseModel


class RawTopicMonitorResult(BaseModel):
	rawTopicErrorCount:int = None
	rawTopicErrorDetails:Dict = None


class PipelineMonitorResult(BaseModel):
	errorCount:int = None
	errorSummary:Dict = {}
	errorDetails:List[Any] = []

class MonitorResult(BaseModel):
	hasError:bool = False
	rawTopicError:RawTopicMonitorResult = None
	pipelineError:PipelineMonitorResult = None




from datetime import datetime

from pydantic import BaseModel
from pydantic.types import Decimal

from watchmen_model.common import Storable


class SQLQueryPerformance(Storable,BaseModel):
	id:str = None
	queryText:str = None
	querySpent:Decimal = None
	queryDate:datetime = None
	resultCount:int = None
	queryParse:dict = None

#
# import sqlparse
#
#
# def parse_sql(query):
# 	parsed = sqlparse.parse(query)[0]
# 	select = []
# 	where = ['WHERE']
# 	condition = None
# 	group_by = ['GROUP BY']
#
# 	# Get SELECT fields
# 	for token in parsed.tokens:
# 		if token.ttype is sqlparse.tokens.Wildcard:
# 			select.append(token.value)
# 		elif token.is_keyword and token.value.upper() == 'FROM':
# 			break
# 		else:
# 			select.append(token.value)
#
# 	# Get WHERE condition
# 	for token in parsed.tokens:
# 		if token.is_keyword and token.value.upper() == 'WHERE':
# 			i = parsed.tokens.index(token) + 1
# 			while parsed.tokens[i].value.upper() != 'GROUP' and parsed.tokens[i].value.upper() != 'ORDER':
# 				where.append(parsed.tokens[i].value)
#
# 				# Set condition
# 				if token.ttype in [sqlparse.tokens.Comparison, sqlparse.tokens.Operator]:
# 					condition = token.value
#
# 				i += 1
#
# 			# Get GROUP BY names
# 	for token in parsed.tokens:
# 		if token.is_keyword and token.value.upper() == 'GROUP':
# 			i = parsed.tokens.index(token) + 2
# 			while parsed.tokens[i].ttype is not sqlparse.tokens.Whitespace:
# 				group_by.append(parsed.tokens[i].value)
# 				i += 1
#
# 	return {
# 		'select': select,
# 		'where': where,
# 		'condition': condition,
# 		'group_by': group_by
# 	}
#
#
# query = "SELECT id, name, age FROM users WHERE age > 30 AND gender = 'M' GROUP BY department, status;"
# result = parse_sql(query)
# print(result)




from .data_result_set import DataResult, DataResultSet, DataResultSetCell, DataResultSetRow
from .graphic import GraphicPosition, GraphicRect, GraphicSize
from .model import DataModel, SettingsModel
from .pagination import DataPage, Pageable, PageDataCell, PageDataRow, PageDataSet
from .parameter_and_condition import ComputedParameter, ConstantParameter, construct_parameter, \
	construct_parameter_condition, construct_parameter_conditions, construct_parameter_joint, Parameter, \
	ParameterComputeType, ParameterCondition, ParameterExpression, ParameterExpressionOperator, ParameterJoint, \
	ParameterJointType, ParameterKind, TopicFactorParameter, VariablePredefineFunctions
from .storable import Auditable, LastVisit, OptimisticLock, Storable
from .tuple import TenantBasedTuple, Tuple, UserBasedTuple
from .tuple_ids import AchievementPluginTaskId, BreakdownTargetId, BucketId, ChangeJsonId, ChangeRecordId, \
	CollectorModelConfigId, CollectorTableConfigId, CompetitiveLockId, ConnectedSpaceId, ConvergenceId, \
	ConvergenceTargetId, ConvergenceTargetVariableMappingId, ConvergenceVariableId, DashboardId, DataSourceId, \
	DerivedObjectiveId, EnumId, EnumItemId, EventDefinitionId, EventTriggerId, ExternalWriterId, FactorId, \
	IndicatorId, ModelTriggerId, NotificationDefinitionId, ObjectiveFactorId, ObjectiveId, ObjectiveTargetId, PatId, \
	PipelineActionId, PipelineGraphicId, PipelineId, PipelineStageId, PipelineUnitId, PluginId, ReportFunnelId, \
	ReportId, ScheduledTaskId, SpaceId, SubjectDatasetColumnId, SubjectId, SubscriptionEventId, \
	SubscriptionEventLockId, TableTriggerId, TenantId, TopicId, UserGroupId, UserId



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import Auditable, BucketId, DataModel, DerivedObjectiveId, FactorId, LastVisit, \
	ObjectiveId, ObjectiveTargetId, SubjectDatasetColumnId, UserBasedTuple, BreakdownTargetId
from watchmen_utilities import ArrayHelper
from .measure_method import MeasureMethod
from .objective import Objective


def construct_objective(definition: Optional[Union[dict, Objective]]) -> Optional[Objective]:
	if definition is None:
		return None
	elif isinstance(definition, Objective):
		return definition
	else:
		# noinspection PyArgumentList
		return Objective(**definition)


class BreakdownDimensionType(str, Enum):
	VALUE = 'value',
	BUCKET = 'bucket',
	TIME_RELATED = 'time'


class BreakdownDimension(DataModel, BaseModel):
	"""
	when type is VALUE, which means no bucket, no time measure method. use the original value as dimension
	"""
	type: BreakdownDimensionType = None
	# if measure on factor, factor id must be given
	factorOrColumnId: Union[FactorId, SubjectDatasetColumnId] = None
	# bucket for any measure on type
	bucketId: BucketId = None
	# only when factor/column is date, and adaptable time measure method could be applied
	# for example, if factor is date, then YEAR/QUARTER/MONTH/etc. could be applied to it
	# if factor is year, then only YEAR could be applied to it.
	timeMeasureMethod: MeasureMethod = None


def construct_breakdown_dimension(dimension: Optional[Union[dict, BreakdownDimension]]) -> Optional[BreakdownDimension]:
	if dimension is None:
		return None
	elif isinstance(dimension, BreakdownDimension):
		return dimension
	else:
		return BreakdownDimension(**dimension)


def construct_breakdown_dimensions(dimensions: Optional[list] = None) -> Optional[List[BreakdownDimension]]:
	if dimensions is None:
		return []
	else:
		return ArrayHelper(dimensions).map(lambda x: construct_breakdown_dimension(x)).to_list()


class BreakdownTarget(DataModel, BaseModel):
	uuid: BreakdownTargetId;
	targetId: ObjectiveTargetId = None
	name: str = None
	dimensions: List[BreakdownDimension] = []

	def __setattr__(self, name, value):
		if name == 'dimensions':
			super().__setattr__(name, construct_breakdown_dimensions(value))
		else:
			super().__setattr__(name, value)


def construct_breakdown_target(target: Optional[Union[dict, BreakdownTarget]]) -> Optional[BreakdownTarget]:
	if target is None:
		return None
	elif isinstance(target, BreakdownTarget):
		return target.dict()
	else:
		return BreakdownTarget(**target)


def construct_breakdown_targets(targets: Optional[list] = None) -> Optional[List[BreakdownTarget]]:
	if targets is None:
		return []
	else:
		return ArrayHelper(targets).map(lambda x: construct_breakdown_target(x)).to_list()


class DerivedObjective(UserBasedTuple, Auditable, LastVisit, BaseModel):
	derivedObjectiveId: DerivedObjectiveId = None
	name: str = None
	description: str = None
	objectiveId: ObjectiveId
	definition: Objective
	breakdownTargets: List[BreakdownTarget] = None

	def __setattr__(self, name, value):
		if name == 'definition':
			super().__setattr__(name, construct_objective(value))
		elif name == 'breakdownTargets':
			super().__setattr__(name, construct_breakdown_targets(value))
		else:
			super().__setattr__(name, value)








from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from watchmen_model.common import BucketId, ConvergenceId, ConvergenceTargetId, ConvergenceTargetVariableMappingId, \
	ConvergenceVariableId, DataModel, ObjectiveId, ObjectiveTargetId, OptimisticLock, TenantBasedTuple, UserGroupId
from watchmen_utilities import ArrayHelper


class ConvergenceVariableType(str, Enum):
	TIMEFRAME = 'timeframe',
	BUCKET = 'bucket',
	FREE_WALK = 'free-walk'


class ConvergenceVariableAxis(str, Enum):
	X = 'x'
	Y = 'y'


class ConvergenceVariable(DataModel, BaseModel):
	uuid: ConvergenceVariableId = None
	type: ConvergenceVariableType = None
	name: str = None
	axis: ConvergenceVariableAxis = None


class ConvergenceTimeFrameVariableKind(str, Enum):
	YEAR = 'year',
	HALF_YEAR = 'half-year',
	QUARTER = 'quarter',
	MONTH = 'month',
	WEEK = 'week',
	DAY = 'day'


class TimeFrameConvergenceVariableValue(DataModel, BaseModel):
	start: str = None
	end: str = None


def construct_timeframe_variable_value(
		target: Optional[Union[dict, TimeFrameConvergenceVariableValue]]
) -> Optional[TimeFrameConvergenceVariableValue]:
	if target is None:
		return None
	elif isinstance(target, TimeFrameConvergenceVariableValue):
		return target
	else:
		# noinspection PyArgumentList
		return TimeFrameConvergenceVariableValue(**target)


def construct_timeframe_variable_values(
		targets: Optional[list] = None) -> Optional[List[TimeFrameConvergenceVariableValue]]:
	if targets is None:
		return None
	else:
		return ArrayHelper(targets).map(lambda x: construct_timeframe_variable_value(x)).to_list()


class ConvergenceTimeFrameVariable(ConvergenceVariable):
	type: ConvergenceVariableType = ConvergenceVariableType.TIMEFRAME
	# use kind and till to compute values
	kind: ConvergenceTimeFrameVariableKind = None
	till: str = None
	times: int = None
	values: List[TimeFrameConvergenceVariableValue] = []

	def __setattr__(self, name, value):
		if name == 'values':
			super().__setattr__(name, construct_timeframe_variable_values(value))
		else:
			super().__setattr__(name, value)


class ConvergenceBucketVariable(ConvergenceVariable):
	type: ConvergenceVariableType = ConvergenceVariableType.BUCKET
	bucketId: BucketId = None


class ConvergenceFreeWalkVariable(ConvergenceVariable):
	type: ConvergenceVariableType = ConvergenceVariableType.FREE_WALK
	values: List[str] = []


CONVERGENCE_TARGET_VARIABLE_MAPPING_IGNORED = '#'


class ConvergenceTargetVariableMapping(DataModel, BaseModel):
	uuid: ConvergenceTargetVariableMappingId = None
	objectiveVariableName: str = None
	#CONVERGENCE_TARGET_VARIABLE_MAPPING_IGNORED
	variableId: ConvergenceVariableId = None


def construct_target_mapping(target: Optional[Union[dict, ConvergenceTargetVariableMapping]]) -> Optional[
	ConvergenceTargetVariableMapping]:
	if target is None:
		return None
	elif isinstance(target, ConvergenceTargetVariableMapping):
		return target
	else:
		# noinspection PyArgumentList
		return ConvergenceTargetVariableMapping(**target)


def construct_target_mappings(targets: Optional[list] = None) -> Optional[List[ConvergenceTargetVariableMapping]]:
	if targets is None:
		return None
	else:
		return ArrayHelper(targets).map(lambda x: construct_target_mapping(x)).to_list()


class ConvergenceTarget(DataModel, BaseModel):
	uuid: ConvergenceTargetId = None
	objectiveId: ObjectiveId = None
	targetId: ObjectiveTargetId = None
	useTimeFrame: bool = None
	mapping: List[ConvergenceTargetVariableMapping] = []
	# starts from 0
	row: int = None
	# starts from 0
	col: int = None

	def __setattr__(self, name, value):
		if name == 'mapping':
			super().__setattr__(name, construct_target_mappings(value))
		else:
			super().__setattr__(name, value)


def construct_target(target: Optional[Union[dict, ConvergenceTarget]]) -> Optional[ConvergenceTarget]:
	if target is None:
		return None
	elif isinstance(target, ConvergenceTarget):
		return target
	else:
		# noinspection PyArgumentList
		return ConvergenceTarget(**target)


def construct_targets(targets: Optional[list] = None) -> Optional[List[ConvergenceTarget]]:
	if targets is None:
		return None
	else:
		return ArrayHelper(targets).map(lambda x: construct_target(x)).to_list()


def construct_variable(variable: Optional[Union[dict, ConvergenceVariable]]) -> Optional[ConvergenceVariable]:
	if variable is None:
		return None
	elif isinstance(variable, ConvergenceVariable):
		return variable
	else:
		type = variable.get('type')
		if type == ConvergenceVariableType.BUCKET:
			return ConvergenceBucketVariable(**variable)
		elif type == ConvergenceVariableType.TIMEFRAME:
			return ConvergenceTimeFrameVariable(**variable)
		elif type == ConvergenceVariableType.FREE_WALK:
			return ConvergenceFreeWalkVariable(**variable)
		else:
			raise Exception(f'Convergence variable type[{type}] cannot be recognized.')


def construct_variables(variables: Optional[list] = None) -> Optional[List[ConvergenceVariable]]:
	if variables is None:
		return None
	else:
		return ArrayHelper(variables).map(lambda x: construct_variable(x)).to_list()


class Convergence(TenantBasedTuple, OptimisticLock, BaseModel):
	convergenceId: ConvergenceId = None
	name: str = None
	description: str = None
	variables: List[ConvergenceVariable] = []
	targets: List[ConvergenceTarget] = []
	groupIds: List[UserGroupId] = None

	def __setattr__(self, name, value):
		if name == 'targets':
			super().__setattr__(name, construct_targets(value))
		elif name == 'variables':
			super().__setattr__(name, construct_variables(value))
		else:
			super().__setattr__(name, value)
